@@ -16,9 +16,18 @@
 /* */
 #undef HAVE_MAXINT_CONVERSION_BUG
 
+/* */
+#undef HAVE_STDLIB_H_X
+
 /* */
 #undef HAVE_SOCKLEN_T
 
+/* */
+#undef HAVE_VOLATILE
+
+/* */
+#undef HAVE_ERRNO
+
 /* */
 #undef C_BIG_ENDIAN
 
@@ -46,27 +55,57 @@
 /* */
 #undef HAVE_STRNCMP
 
+/* */
+#undef HAVE_STRCASECMP
+
+/* */
+#undef HAVE_STRNCASECMP
+
 /* */
 #undef HAVE_STRCSPN
 
 /* */
 #undef HAVE_STRSTR
 
+/* */
+#undef HAVE_MEMCMP
+
+/* */
+#undef HAVE_MEMCHR
+
+/* */
+#undef HAVE_MEMCPY
+
 /* */
 #undef HAVE_MEMMOVE
 
+/* */
+#undef HAVE_MEMSET
+
 /* */
 #undef HAVE_STRERROR
 
 /* */
 #undef HAVE_SIGSETJMP
 
+/* */
+#undef HAVE_POW
+
+/* */
+#undef HAVE_POWF
+
 /* */
 #undef HAVE_GETHOSTBYNAME
 
 /* */
 #undef HAVE_GETHOSTBYNAME_BUG
 
+/* */
+#undef SUPPORT_IPV6
+
+/* */
+#undef SUPPORT_IPV6_SCOPE
+
 /* */
 #undef DEBUGLEVEL
 
@@ -94,6 +133,9 @@
 /* */
 #undef X2
 
+/* */
+#undef HAVE_XSETLOCALE
+
 /* */
 #undef HAVE_SSL
 
@@ -109,6 +151,9 @@
 /* */
 #undef HAVE_BZIP2
 
+/* */
+#undef HAVE_LZMA
+
 /* */
 #undef G
 
@@ -133,6 +178,9 @@
 /* */
 #undef GRDRV_ATHEOS
 
+/* */
+#undef GRDRV_GRX
+
 /* */
 #undef DONT_INCLUDE_SETJMP
 
