@@ -14,7 +14,7 @@ static struct list_head image_cache = { &image_cache, &image_cache };
 
 /* xyw_meaning either MEANING_DIMS or MEANING_AUTOSCALE. */
 struct cached_image *find_cached_image(int bg, unsigned char *url, int xw, int
-		yw, int xyw_meaning, int scale, int aspect)
+		yw, int xyw_meaning, int scale, unsigned aspect)
 {
 	struct cached_image *i;
 	if (xw>=0&&yw>=0&&xyw_meaning==MEANING_DIMS){
@@ -23,7 +23,7 @@ struct cached_image *find_cached_image(int bg, unsigned char *url, int xw, int
 		 */
 		foreach (i, image_cache) {
 			if (i->background_color == bg
-				&& !strcmp(i->url, url)
+				&& !strcmp(cast_const_char i->url, cast_const_char url)
 				&& i->wanted_xw==xw
 				&& i->wanted_yw==yw
 				&& i->wanted_xyw_meaning==xyw_meaning
@@ -32,7 +32,7 @@ struct cached_image *find_cached_image(int bg, unsigned char *url, int xw, int
 	}else{
 		foreach (i, image_cache) {
 			if (i->background_color == bg
-				&& !strcmp(i->url, url)
+				&& !strcmp(cast_const_char i->url, cast_const_char url)
 				&& i->wanted_xw==xw
 				&& i->wanted_yw==yw
 				&& i->wanted_xyw_meaning==xyw_meaning 
@@ -110,10 +110,10 @@ static int shrink_image_cache(int u)
 	return r | (list_empty(image_cache) ? ST_CACHE_EMPTY : 0);
 }
 
-unsigned long imgcache_info(int type)
+my_uintptr_t imgcache_info(int type)
 {
 	struct cached_image *i;
-	unsigned long n = 0;
+	my_uintptr_t n = 0;
 	foreach(i, image_cache) {
 		switch (type) {
 			case CI_BYTES:
@@ -134,7 +134,7 @@ unsigned long imgcache_info(int type)
 
 void init_imgcache(void)
 {
-	register_cache_upcall(shrink_image_cache, "imgcache");
+	register_cache_upcall(shrink_image_cache, MF_GPI, cast_uchar "imgcache");
 }
 
 #endif
