@@ -8,15 +8,18 @@
 static void in_term(struct terminal *);
 static void check_if_no_terminal(void);
 
-
 int hard_write(int fd, unsigned char *p, int l)
 {
-	int w = 1;
 	int t = 0;
-	while (l > 0 && w) {
-		EINTRLOOP(w, write(fd, p, l));
+	while (l > 0) {
+		int w;
+		EINTRLOOP(w, (int)write(fd, p, l));
 		if (w < 0)
 			return -1;
+		if (!w) {
+			errno = ENOSPC;
+			break;
+		}
 		t += w;
 		p += w;
 		l -= w;
@@ -29,7 +32,7 @@ int hard_read(int fd, unsigned char *p, int l)
 	int r = 1;
 	int t = 0;
 	while (l > 0 && r) {
-		EINTRLOOP(r, read(fd, p, l));
+		EINTRLOOP(r, (int)read(fd, p, l));
 		if (r < 0)
 			return -1;
 		t += r;
@@ -46,10 +49,10 @@ unsigned char *get_cwd(void)
 	unsigned char *gcr;
 	while (1) {
 		buf = mem_alloc(bufsize);
-		ENULLLOOP(gcr, getcwd(buf, bufsize));
+		ENULLLOOP(gcr, cast_uchar getcwd(cast_char buf, bufsize));
 		if (gcr) return buf;
 		mem_free(buf);
-		if (errno != ERANGE) return NULL;
+		if (errno != ERANGE) break;
 		if ((unsigned)bufsize > MAXINT - 128) overalloc();
 		bufsize += 128;
 	}
@@ -60,7 +63,12 @@ void set_cwd(unsigned char *path)
 {
 	int rs;
 	if (path)
-		EINTRLOOP(rs, chdir(path));
+		EINTRLOOP(rs, chdir(cast_const_char path));
+}
+
+unsigned char get_attribute(int fg, int bg)
+{
+	return ((bg & 7) << 3) | (fg & 7) | ((fg & 8) << 3);
 }
 
 struct list_head terminals = {&terminals, &terminals};
@@ -69,6 +77,8 @@ static void alloc_term_screen(struct terminal *term, int x, int y)
 {
 	chr *s, *t;
 	NO_GFX;
+	if (x < 0) x = 1;
+	if (y < 0) y = 1;
 	if (x && (unsigned)x * (unsigned)y / (unsigned)x != (unsigned)y) overalloc();
 	if ((unsigned)x * (unsigned)y > MAXINT / sizeof(*term->screen)) overalloc();
 	s = mem_realloc(term->screen, x * y * sizeof(*term->screen));
@@ -121,13 +131,13 @@ static void redraw_terminal_ev(struct terminal *term, int e)
 	term->redrawing = 0;
 }
 
-void redraw_terminal(struct terminal *term)
+static void redraw_terminal(struct terminal *term)
 {
 	NO_GFX;
 	redraw_terminal_ev(term, EV_REDRAW);
 }
 
-void redraw_terminal_all(struct terminal *term)
+static void redraw_terminal_all(struct terminal *term)
 {
 	NO_GFX;
 	redraw_terminal_ev(term, EV_RESIZE);
@@ -138,12 +148,12 @@ static void erase_screen(struct terminal *term)
 	NO_GFX;
 	if (!term->master || !is_blocked()) {
 		if (term->master) want_draw();
-		hard_write(term->fdout, "\033[2J\033[1;1H", 10);
+		hard_write(term->fdout, cast_uchar "\033[2J\033[1;1H", 10);
 		if (term->master) done_draw();
 	}
 }
 
-void redraw_terminal_cls(struct terminal *term)
+static void redraw_terminal_cls(struct terminal *term)
 {
 	NO_GFX;
 	erase_screen(term);
@@ -324,7 +334,11 @@ void draw_to_window(struct window *win, void (*fn)(struct terminal *term, void *
 		memcpy(&r1, &term->dev->clip, sizeof(struct rect));
 		for (i = 0; i < s->m; i++) if (is_rect_valid(r = &s->r[i])) {
 			drv->set_clip_area(term->dev, r);
-			pr(fn(term, data)) return;
+			pr(fn(term, data)) {
+#ifdef OOPS
+				return;
+#endif
+			}
 			a = 1;
 		}
 		if (!a) {
@@ -497,17 +511,17 @@ static struct term_spec *get_term_spec(unsigned char *term)
 {
 	struct term_spec *t;
 	NO_GFX;
-	foreach(t, term_specs) if (!strcasecmp(t->term, term)) return t;
+	foreach(t, term_specs) if (!strcasecmp(cast_const_char t->term, cast_const_char term)) return t;
 	return &dumb_term;
 }
 
 struct term_spec *new_term_spec(unsigned char *term)
 {
 	struct term_spec *t;
-	foreach(t, term_specs) if (!strcasecmp(t->term, term)) return t;
+	foreach(t, term_specs) if (!strcasecmp(cast_const_char t->term, cast_const_char term)) return t;
 	t = mem_alloc(sizeof(struct term_spec));
 	memcpy(t, &dumb_term, sizeof(struct term_spec));
-	if (strlen(term) < MAX_TERM_LEN) strcpy(t->term, term);
+	if (strlen(cast_const_char term) < MAX_TERM_LEN) strcpy(cast_char t->term, cast_const_char term);
 	else memcpy(t->term, term, MAX_TERM_LEN - 1), t->term[MAX_TERM_LEN - 1] = 0;
 	add_to_list(term_specs, t);
 	sync_term_specs();
@@ -562,9 +576,9 @@ static int process_utf_8(struct terminal *term, struct event *ev)
 		unsigned char *p;
 		unsigned c;
 		if (ev->x <= 0 || ev->x >= 0x100) goto direct;
-		if ((l = strlen(term->utf8_buffer)) >= sizeof(term->utf8_buffer) - 1 || ev->x < 0x80 || ev->x >= 0xc0)
+		if ((l = strlen(cast_const_char term->utf8_buffer)) >= sizeof(term->utf8_buffer) - 1 || ev->x < 0x80 || ev->x >= 0xc0)
 			term->utf8_buffer[0] = 0, l = 0;
-		term->utf8_buffer[l] = ev->x;
+		term->utf8_buffer[l] = (unsigned char)ev->x;
 		term->utf8_buffer[l + 1] = 0;
 		p = term->utf8_buffer;
 		GET_UTF_8(p, c);
@@ -604,7 +618,7 @@ struct terminal *init_gfx_term(void (*root_window)(struct window *, struct event
 	term->y = dev->size.y2;
 	term->last_mouse_x = term->last_mouse_y = term->last_mouse_b = MAXINT;
 	term->environment = !(drv->flags & GD_ONLY_1_WINDOW) ? ENV_G : 0;
-	if (!strcasecmp(drv->name, "x")) term->environment |= ENV_XWIN;
+	if (!strcasecmp(cast_const_char drv->name, "x")) term->environment |= ENV_XWIN;
 	term->spec = &gfx_term;
 	if ((cwd = get_cwd())) {
 		safe_strncpy(term->cwd, cwd, MAX_CWD_LEN);
@@ -723,7 +737,7 @@ static void in_term(struct terminal *term)
 	if ((unsigned)term->qlen + ALLOC_GR > MAXINT) overalloc();
 	iq = mem_realloc(term->input_queue, term->qlen + ALLOC_GR);
 	term->input_queue = iq;
-	EINTRLOOP(r, read(term->fdin, iq + term->qlen, ALLOC_GR));
+	EINTRLOOP(r, (int)read(term->fdin, iq + term->qlen, ALLOC_GR));
 	if (r <= 0) {
 		if (r == -1 && errno != ECONNRESET) error("ERROR: error %d on terminal: could not read event", errno);
 		destroy_terminal(term);
@@ -748,8 +762,8 @@ static void in_term(struct terminal *term)
 		memcpy(term->cwd, iq + sizeof(struct event) + MAX_TERM_LEN, MAX_CWD_LEN);
 		term->cwd[MAX_CWD_LEN - 1] = 0;
 		term->environment = *(int *)(iq + sizeof(struct event) + MAX_TERM_LEN + MAX_CWD_LEN);
-		ev->b = (long)(iq + sizeof(struct event) + MAX_TERM_LEN + MAX_CWD_LEN + sizeof(int));
-		r = sizeof(struct event) + MAX_TERM_LEN + MAX_CWD_LEN + 2 * sizeof(int) + init_len;
+		ev->b = (my_intptr_t)(iq + sizeof(struct event) + MAX_TERM_LEN + MAX_CWD_LEN + sizeof(int));
+		r = (int)sizeof(struct event) + MAX_TERM_LEN + MAX_CWD_LEN + 2 * (int)sizeof(int) + init_len;
 		sync_term_specs();
 	}
 	if (ev->ev == EV_REDRAW || ev->ev == EV_RESIZE || ev->ev == EV_INIT) {
@@ -795,8 +809,8 @@ static inline int getcompcode(int c)
 	return (c<<1 | (c&4)>>2) & 7;
 }
 
-unsigned char frame_dumb[48] =	"   ||||++||++++++--|-+||++--|-+----++++++++     ";
-static unsigned char frame_vt100[48] =	"aaaxuuukkuxkjjjkmvwtqnttmlvwtqnvvwwmmllnnjla    ";
+unsigned char frame_dumb[49] =	"   ||||++||++++++--|-+||++--|-+----++++++++     ";
+static unsigned char frame_vt100[49] =	"aaaxuuukkuxkjjjkmvwtqnttmlvwtqnvvwwmmllnnjla    ";
 static unsigned char frame_koi[48] = {
 	144,145,146,129,135,178,180,167,
 	166,181,161,168,174,173,172,131,
@@ -847,11 +861,12 @@ static inline char_t utf8_hack(char_t c)
 
 #define SETPOS(x, y)							\
 {									\
-	add_to_str(&a, &l, "\033[");					\
+	add_to_str(&a, &l, cast_uchar "\033[");				\
 	add_num_to_str(&a, &l, (y) + 1);				\
-	add_to_str(&a, &l, ";");					\
+	add_to_str(&a, &l, cast_uchar ";");				\
 	add_num_to_str(&a, &l, (x) + 1);				\
-	add_to_str(&a, &l, "H");					\
+	add_to_str(&a, &l, cast_uchar "H");				\
+	n_chars = 0;							\
 }
 
 #define PRINT_CHAR(p)							\
@@ -862,8 +877,8 @@ static inline char_t utf8_hack(char_t c)
 	if (s->mode == TERM_LINUX) {					\
 		if (s->m11_hack) {					\
 			if (frm != mode) {				\
-				if (!(mode = frm)) add_to_str(&a, &l, "\033[10m");\
-				else add_to_str(&a, &l, "\033[11m");	\
+				if (!(mode = frm)) add_to_str(&a, &l, cast_uchar "\033[10m");\
+				else add_to_str(&a, &l, cast_uchar "\033[11m");\
 			}						\
 		}							\
 		if (s->restrict_852 && frm && c >= 176 && c < 224) {	\
@@ -871,8 +886,8 @@ static inline char_t utf8_hack(char_t c)
 		}							\
 	} else if (s->mode == TERM_VT100) {				\
 		if (frm != mode) {					\
-			if (!(mode = frm)) add_to_str(&a, &l, "\017");	\
-			else add_to_str(&a, &l, "\016");		\
+			if (!(mode = frm)) add_to_str(&a, &l, cast_uchar "\017");\
+			else add_to_str(&a, &l, cast_uchar "\016");	\
 		}							\
 		if (frm && c >= 176 && c < 224) c = frame_vt100[c - 176];\
 	} else if (s->mode == TERM_KOI8 && frm && c >= 176 && c < 224) { c = frame_koi[c - 176];\
@@ -881,7 +896,7 @@ static inline char_t utf8_hack(char_t c)
 	if (!(A & 0100) && (A >> 3) == (A & 7)) A = (A & 070) | 7 * !(A & 020);\
 	if (A != attrib) {						\
 		attrib = A;						\
-		add_to_str(&a, &l, "\033[0");				\
+		add_to_str(&a, &l, cast_uchar "\033[0");		\
 		if (s->col) {						\
 			unsigned char m[4];				\
 			m[0] = ';'; m[1] = '3'; m[3] = 0;		\
@@ -891,13 +906,13 @@ static inline char_t utf8_hack(char_t c)
 			m[2] = ((attrib >> 3) & 7) + '0';		\
 			add_to_str(&a, &l, m);				\
 		} else if (getcompcode(attrib & 7) < getcompcode(attrib >> 3 & 7))\
-			add_to_str(&a, &l, ";7");			\
-		if (attrib & 0100) add_to_str(&a, &l, ";1");		\
-		add_to_str(&a, &l, "m");				\
+			add_to_str(&a, &l, cast_uchar ";7");		\
+		if (attrib & 0100) add_to_str(&a, &l, cast_uchar ";1");	\
+		add_to_str(&a, &l, cast_uchar "m");			\
 	}								\
-	if (c >= ' ' && c != 127/* && c != 155*/) {			\
+	if (c >= ' ' && c != 127 /*&& c != 155*/) {			\
 		if (c < 128 || frm || term->spec->charset != utf8_table) {\
-			add_chr_to_str(&a, &l, c);			\
+			add_chr_to_str(&a, &l, (unsigned char)c);	\
 		} else {						\
 		/*							\
 		 * Linux UTF-8 console is broken and doesn't advance cursor\
@@ -908,7 +923,7 @@ static inline char_t utf8_hack(char_t c)
 			unsigned char *r;				\
 			c = utf8_hack(c);				\
 			r = u2cp(c, 0, 1);				\
-			if (!(r && r[0] >= 32 && r[0] < 127 && !r[1])) r = "*";	\
+			if (!(r && r[0] >= 32 && r[0] < 127 && !r[1])) r = cast_uchar "*";\
 			add_chr_to_str(&a, &l, r[0]);			\
 			if (cx + 1 < term->x)				\
 				add_chr_to_str(&a, &l, 8);		\
@@ -921,7 +936,8 @@ static inline char_t utf8_hack(char_t c)
 	else if (!c || c == 1) add_chr_to_str(&a, &l, ' ');		\
 	else add_chr_to_str(&a, &l, '.');				\
 	cx++;								\
-}									\
+	n_chars++;							\
+}
 
 void redraw_all_terminals(void)
 {
@@ -933,6 +949,7 @@ void redraw_screen(struct terminal *term)
 {
 	int x, y, p = 0;
 	int cx = term->lcx, cy = term->lcy;
+	unsigned n_chars = MAXINT / 2;
 	unsigned char *a;
 	int attrib = -1;
 	int mode = -1;
@@ -942,13 +959,27 @@ void redraw_screen(struct terminal *term)
 	if (!term->dirty || (term->master && is_blocked())) return;
 	a = init_str();
 	s = term->spec;
-	for (y = 0; y < term->y; y++)
+	for (y = 0; y < term->y; y++) {
+		if (!memcmp(&term->screen[p], &term->last_screen[p], sizeof(chr) * term->x)) {
+			p += term->x;
+			continue;
+		}
 		for (x = 0; x < term->x; x++, p++) {
 			int i;
 			if (y == term->y - 1 && x == term->x - 1) break;
-			if (!memcmp(&term->screen[p], &term->last_screen[p], sizeof(chr))) continue;
+			if (term->screen[p].ch == term->last_screen[p].ch && term->screen[p].at == term->last_screen[p].at) {
+#if defined(ENABLE_UTF8) && !defined(HAVE_ATTR_PACKED)
+				/* make sure that padding is identical */
+				memcpy(&term->last_screen[p], &term->screen[p], sizeof(chr));
+#endif
+				continue;
+			}
 			/*if ((term->screen[p].at & 0x38) == (term->last_screen[p].at & 0x38) && (term->screen[p].ch == 0 || term->screen[p].ch == 1 || term->screen[p].ch == ' ') && (term->last_screen[p].ch == 0 || term->last_screen[p].ch == 1 || term->last_screen[p].ch == ' ') && (x != term->cx || y != term->cy)) continue;*/
+			/*fprintf(stderr, "%d.%d : %d-%d -> %d-%d\n", x, y, term->last_screen[p].ch, term->last_screen[p].at, term->screen[p].ch, term->screen[p].at);*/
 			memcpy(&term->last_screen[p], &term->screen[p], sizeof(chr));
+#ifdef OPENVMS
+			if (n_chars >= 16) cy = -1;
+#endif
 			if (cx == x && cy == y) goto pc;/*PRINT_CHAR(p)*/
 			else if (cy == y && x - cx < 10 && x - cx > 0) {
 				for (i = x - cx; i >= 0; i--) {
@@ -963,22 +994,23 @@ void redraw_screen(struct terminal *term)
 				goto ppc;
 			}
 		}
+	}
 	if (l) {
-		if (s->col) add_to_str(&a, &l, "\033[37;40m");
-		add_to_str(&a, &l, "\033[0m");
-		if (s->mode == TERM_LINUX && s->m11_hack) add_to_str(&a, &l, "\033[10m");
-		if (s->mode == TERM_VT100) add_to_str(&a, &l, "\017");
+		if (s->col) add_to_str(&a, &l, cast_uchar "\033[37;40m");
+		add_to_str(&a, &l, cast_uchar "\033[0m");
+		if (s->mode == TERM_LINUX && s->m11_hack) add_to_str(&a, &l, cast_uchar "\033[10m");
+		if (s->mode == TERM_VT100) add_to_str(&a, &l, cast_uchar "\017");
 	}
 	term->lcx = cx;
 	term->lcy = cy;
 	if (term->cx != term->lcx || term->cy != term->lcy) {
 		term->lcx = term->cx;
 		term->lcy = term->cy;
-		add_to_str(&a, &l, "\033[");
+		add_to_str(&a, &l, cast_uchar "\033[");
 		add_num_to_str(&a, &l, term->cy + 1);
-		add_to_str(&a, &l, ";");
+		add_to_str(&a, &l, cast_uchar ";");
 		add_num_to_str(&a, &l, term->cx + 1);
-		add_to_str(&a, &l, "H");
+		add_to_str(&a, &l, cast_uchar "H");
 	}
 	if (l && term->master) want_draw();
 	hard_write(term->fdout, a, l);
@@ -1011,13 +1043,9 @@ void destroy_terminal(struct terminal *term)
 		} else {
 			unhandle_terminal_signals(term);
 			free_all_itrms();
-#ifndef NO_FORK_ON_EXIT
 			if (!list_empty(terminals)) {
-				pid_t rp;
-				EINTRLOOP(rp, fork());
-				if (rp > 0) _exit(0);
+				os_detach_console();
 			}
-#endif
 		}
 #ifdef G
 	} else {
@@ -1040,7 +1068,7 @@ static void check_if_no_terminal(void)
 	terminate_loop = 1;
 }
 
-void set_char(struct terminal *t, int x, int y, unsigned ch, unsigned at)
+void set_char(struct terminal *t, int x, int y, unsigned ch, unsigned char at)
 {
 	NO_GFX;
 	t->dirty = 1;
@@ -1054,6 +1082,12 @@ void set_char(struct terminal *t, int x, int y, unsigned ch, unsigned at)
 chr *get_char(struct terminal *t, int x, int y)
 {
 	NO_GFX;
+	if (!t->x || !t->y) {
+		static chr empty;
+		empty.ch = ' ';
+		empty.at = 070;
+		return &empty;
+	}
 	if (x >= t->x) x = t->x - 1;
 	if (x < 0) x = 0;
 	if (y >= t->y) y = t->y - 1;
@@ -1061,14 +1095,14 @@ chr *get_char(struct terminal *t, int x, int y)
 	return &t->screen[x + t->x * y];
 }
 
-void set_color(struct terminal *t, int x, int y, unsigned c)
+void set_color(struct terminal *t, int x, int y, unsigned char c)
 {
 	NO_GFX;
 	t->dirty = 1;
 	if (x >= 0 && x < t->x && y >= 0 && y < t->y) t->screen[x + t->x * y].at = (t->screen[x + t->x * y].at & ATTR_FRAME) | (c & ~ATTR_FRAME);
 }
 
-void set_only_char(struct terminal *t, int x, int y, unsigned ch, unsigned at)
+void set_only_char(struct terminal *t, int x, int y, unsigned ch, unsigned char at)
 {
 	chr *cc;
 	NO_GFX;
@@ -1093,7 +1127,7 @@ void set_line(struct terminal *t, int x, int y, int l, chr *line)
 	memcpy(cc, line, i * sizeof(chr));
 }
 
-void set_line_color(struct terminal *t, int x, int y, int l, unsigned c)
+void set_line_color(struct terminal *t, int x, int y, int l, unsigned char c)
 {
 	int i;
 	NO_GFX;
@@ -1103,27 +1137,35 @@ void set_line_color(struct terminal *t, int x, int y, int l, unsigned c)
 		t->screen[x+i + t->x * y].at = (t->screen[x+i + t->x * y].at & ATTR_FRAME) | (c & ~ATTR_FRAME);
 }
 
-void fill_area(struct terminal *t, int x, int y, int xw, int yw, unsigned ch, unsigned at)
+void fill_area(struct terminal *t, int x, int y, int xw, int yw, unsigned ch, unsigned char at)
 {
-	int i,j;
+	int i;
+	chr *p, *ps;
 	NO_GFX;
+	if (x < 0) xw += x, x = 0;
+	if (x + xw > t->x) xw = t->x - x;
+	if (xw <= 0) return;
+	if (y < 0) yw += y, y = 0;
+	if (y + yw > t->y) yw = t->y - y;
+	if (yw <= 0) return;
 	t->dirty = 1;
-	for (j = y >= 0 ? 0 : -y; j < yw && y+j < t->y; j++) {
-		chr *cc;
-		i = x >= 0 ? 0 : -x;
-		cc = &t->screen[x+i + t->x*(y+j)];
-		for (; i < xw && x+i < t->x; i++) {
-			cc->ch = ch;
-			cc->at = at;
-			cc++;
-		}
+	p = ps = &t->screen[x + t->x * y];
+	for (i = 0; i < xw; i++) {
+		p->ch = ch;
+		p->at = at;
+		p++;
+	}
+	p = ps;
+	for (i = 1; i < yw; i++) {
+		p += t->x;
+		memcpy(p, ps, xw * sizeof(chr));
 	}
 }
 
 static int p1[] = { 218, 191, 192, 217, 179, 196 };
 static int p2[] = { 201, 187, 200, 188, 186, 205 };
 
-void draw_frame(struct terminal *t, int x, int y, int xw, int yw, unsigned c, int w)
+void draw_frame(struct terminal *t, int x, int y, int xw, int yw, unsigned char c, int w)
 {
 	int *p = w > 1 ? p2 : p1;
 	NO_GFX;
@@ -1138,7 +1180,7 @@ void draw_frame(struct terminal *t, int x, int y, int xw, int yw, unsigned c, in
 	fill_area(t, x+1, y+yw-1, xw-2, 1, p[5], c);
 }
 
-void print_text(struct terminal *t, int x, int y, int l, unsigned char *text, unsigned c)
+void print_text(struct terminal *t, int x, int y, int l, unsigned char *text, unsigned char c)
 {
 	NO_GFX;
 	for (; l--; x++) {
@@ -1164,18 +1206,18 @@ void set_cursor(struct terminal *term, int x, int y, int altx, int alty)
 void exec_thread(unsigned char *path, int p)
 {
 	int rs;
-#if defined(HAVE_SETPGID) && !defined(BEOS) && !defined(HAVE_BEGINTHREAD)
+#if defined(HAVE_SETPGID) && !defined(EXEC_IN_THREADS)
 	if (path[0] == 2)
 		EINTRLOOP(rs, setpgid(0, 0));
 #endif
 	exe(path + 1, path[0]);
-	if (path[1 + strlen(path + 1) + 1])
-		EINTRLOOP(rs, unlink(path + 1 + strlen(path + 1) + 1));
+	if (path[1 + strlen(cast_const_char(path + 1)) + 1])
+		EINTRLOOP(rs, unlink(cast_const_char(path + 1 + strlen(cast_const_char(path + 1)) + 1)));
 }
 
 void close_handle(void *p)
 {
-	long h = (my_intptr_t)p;
+	int h = (int)(my_intptr_t)p;
 	int rs;
 	EINTRLOOP(rs, close(h));
 	set_handlers(h, NULL, NULL, NULL, NULL);
@@ -1188,7 +1230,10 @@ static void unblock_terminal(struct terminal *term)
 	if (!F) {
 		set_handlers(term->fdin, (void (*)(void *))in_term, NULL, (void (*)(void *))destroy_terminal, term);
 		unblock_itrm(term->fdin);
-		redraw_terminal_cls(term);
+		/* clear the dirty flag because unblock_itrm queued a resize
+		   event - so avoid double redraw */
+		term->dirty = 0;
+		/*redraw_terminal_cls(term);*/
 #ifdef G
 	} else {
 		drv->unblock(term->dev);
@@ -1196,39 +1241,58 @@ static void unblock_terminal(struct terminal *term)
 	}
 }
 
-void exec_on_terminal(struct terminal *term, unsigned char *path, unsigned char *delete, int fg)
+#ifdef G
+int have_extra_exec(void)
+{
+#ifdef NO_FG_EXEC
+	return 0;
+#else
+	return F && drv->exec;
+#endif
+}
+#endif
+
+void exec_on_terminal(struct terminal *term, unsigned char *path, unsigned char *delet, unsigned char fg)
 {
 	int rs;
 	if (path && !*path) return;
-	if (!path) path="";
+	if (!path) path = cast_uchar "";
 #ifdef NO_FG_EXEC
 	fg = 0;
+#endif
+#ifdef HAVE_EXE_ON_BACKGROUND
+	if (*path) {
+		rs = exe_on_background(path, delet);
+		if (!rs) return;
+	}
 #endif
 	if (term->master) {
 		if (!*path) {
-			if (!F) dispatch_special(delete);
+			if (!F) dispatch_special(delet);
 		} else {
-			long blockh;
+			int blockh;
 			unsigned char *param;
-			if ((!F ? is_blocked() : term->blocked != -1) && fg) {
-				EINTRLOOP(rs, unlink(delete));
+			if (is_blocked() && fg) {
+				if (*delet)
+					EINTRLOOP(rs, unlink(cast_const_char delet));
 				return;
 			}
-			param = mem_alloc(strlen(path) + strlen(delete) + 3);
+			param = mem_alloc(strlen(cast_const_char path) + strlen(cast_const_char delet) + 3);
 			param[0] = fg;
-			strcpy(param + 1, path);
-			strcpy(param + 1 + strlen(path) + 1, delete);
+			strcpy(cast_char(param + 1), cast_const_char path);
+			strcpy(cast_char(param + 1 + strlen(cast_const_char path) + 1), cast_const_char delet);
 			if (fg == 1) {
 				if (!F) block_itrm(term->fdin);
 #ifdef G
 				else if (drv->block(term->dev)) {
 					mem_free(param);
-					EINTRLOOP(rs, unlink(delete));
+					if (*delet)
+						EINTRLOOP(rs, unlink(cast_const_char delet));
 					return;
 				}
 #endif
 			}
-			if ((blockh = start_thread((void (*)(void *, int))exec_thread, param, strlen(path) + strlen(delete) + 3)) == -1) {
+			if ((blockh = start_thread((void (*)(void *, int))exec_thread, param, (int)strlen(cast_const_char path) + (int)strlen(cast_const_char delet) + 3)) == -1) {
 				if (fg == 1) {
 					if (!F) unblock_itrm(term->fdin);
 #ifdef G
@@ -1239,23 +1303,31 @@ void exec_on_terminal(struct terminal *term, unsigned char *path, unsigned char
 				return;
 			}
 			mem_free(param);
-			if (fg == 1) {
+			if (fg == 1
+#ifdef G
+				&& !have_extra_exec()
+#endif
+				) {
 				term->blocked = blockh;
+#ifdef DOS
+				unblock_terminal(term);
+#else
 				set_handlers(blockh, (void (*)(void *))unblock_terminal, NULL, (void (*)(void *))unblock_terminal, term);
 				if (!F) set_handlers(term->fdin, NULL, NULL, (void (*)(void *))destroy_terminal, term);
 				/*block_itrm(term->fdin);*/
+#endif
 			} else {
-				set_handlers(blockh, close_handle, NULL, close_handle, (void *)blockh);
+				set_handlers(blockh, close_handle, NULL, close_handle, (void *)(my_intptr_t)blockh);
 			}
 		}
 	} else {
 		unsigned char *data;
-		data = mem_alloc(strlen(path) + strlen(delete) + 4);
+		data = mem_alloc(strlen(cast_const_char path) + strlen(cast_const_char delet) + 4);
 		data[0] = 0;
 		data[1] = fg;
-		strcpy(data + 2, path);
-		strcpy(data + 3 + strlen(path), delete);
-		hard_write(term->fdout, data, strlen(path) + strlen(delete) + 4);
+		strcpy(cast_char(data + 2), cast_const_char path);
+		strcpy(cast_char(data + 3 + strlen(cast_const_char path)), cast_const_char delet);
+		hard_write(term->fdout, data, (int)strlen(cast_const_char path) + (int)strlen(cast_const_char delet) + 4);
 		mem_free(data);
 	}
 }
@@ -1264,25 +1336,22 @@ void do_terminal_function(struct terminal *term, unsigned char code, unsigned ch
 {
 	unsigned char *x_data;
 	NO_GFX;
-	x_data = mem_alloc(strlen(data) + 2);
+	x_data = mem_alloc(strlen(cast_const_char data) + 2);
 	x_data[0] = code;
-	strcpy(x_data + 1, data);
+	strcpy(cast_char(x_data + 1), cast_const_char data);
 	exec_on_terminal(term, NULL, x_data, 0);
 	mem_free(x_data);
 }
 
 void set_terminal_title(struct terminal *term, unsigned char *title)
 {
-	int i;
-	for (i = 0; i < 10000; i++) if (!title[i]) goto s;
-	title[10000] = 0;
-	s:
-	if (strchr(title, 1)) {
+	if (strlen(cast_const_char title) > 10000) title[10000] = 0;
+	if (strchr(cast_const_char title, 1)) {
 		unsigned char *a, *b;
 		for (a = title, b = title; *a; a++) if (*a != 1) *b++ = *a;
 		*b = 0;
 	}
-	if (term->title && !strcmp(title, term->title)) goto ret;
+	if (term->title && !strcmp(cast_const_char title, cast_const_char term->title)) goto ret;
 	if (term->title) mem_free(term->title);
 	term->title = stracpy(title);
 #ifdef SET_WINDOW_TITLE_UTF_8
@@ -1290,7 +1359,7 @@ void set_terminal_title(struct terminal *term, unsigned char *title)
 		struct conv_table *table;
 		mem_free(title);
 		table = get_translation_table(term->spec->charset, utf8_table);
-		title = convert_string(table, term->title, strlen(term->title), NULL);
+		title = convert_string(table, term->title, strlen(cast_const_char term->title), NULL);
 	}
 #endif
 	if (!F) do_terminal_function(term, TERM_FN_TITLE, title);
